#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for thickener unit example.
"""

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    assert_optimal_termination,
)

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    MomentumBalanceType,
)

from idaes.models.unit_models.separator import SplittingType

from pyomo.environ import (
    units,
)

from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
import idaes.core.util.scaling as iscale
from idaes.core.util.testing import (
    initialization_tester,
)

from idaes.core.util.exceptions import (
    ConfigurationError,
)
from watertap.unit_models.thickener import Thickener, ActivatedSludgeModelType
from watertap.property_models.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
)

from watertap.property_models.activated_sludge.asm2d_properties import (
    ASM2dParameterBlock,
)
from watertap.property_models.activated_sludge.modified_asm2d_properties import (
    ModifiedASM2dParameterBlock,
)
from pyomo.util.check_units import assert_units_consistent
from watertap.costing import WaterTAPCosting
from idaes.core import UnitModelCostingBlock

__author__ = "Alejandro Garciadiego, Adam Atia"

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = Thickener(property_package=m.fs.props)

    assert len(m.fs.unit.config) == 16

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.activated_sludge_model == ActivatedSludgeModelType.ASM1
    assert "underflow" in m.fs.unit.config.outlet_list
    assert "overflow" in m.fs.unit.config.outlet_list
    assert SplittingType.componentFlow is m.fs.unit.config.split_basis


@pytest.mark.unit
def test_list_error():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    with pytest.raises(
        ConfigurationError,
        match="fs.unit encountered unrecognised "
        "outlet_list. This should not "
        "occur - please use overflow "
        "and underflow as outlets.",
    ):
        m.fs.unit = Thickener(
            property_package=m.fs.props, outlet_list=["outlet1", "outlet2"]
        )


# -----------------------------------------------------------------------------
class TestThickASM1(object):
    @pytest.fixture(scope="class")
    def tu(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ASM1ParameterBlock()

        m.fs.unit = Thickener(property_package=m.fs.props)

        m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)

        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(28.0643 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(0.67336 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(3036.2175 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(63.2392 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(
            4442.8377 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(332.5958 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(1922.8108 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1.3748 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(9.1948 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(0.15845 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.55943 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(4.7411 * units.mg / units.liter)
        m.fs.unit.inlet.alkalinity.fix(4.5646 * units.mol / units.m**3)

        m.fs.unit.hydraulic_retention_time.fix()
        m.fs.unit.diameter.fix()

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, tu):

        assert hasattr(tu.fs.unit, "inlet")
        assert len(tu.fs.unit.inlet.vars) == 5
        assert hasattr(tu.fs.unit.inlet, "flow_vol")
        assert hasattr(tu.fs.unit.inlet, "conc_mass_comp")
        assert hasattr(tu.fs.unit.inlet, "temperature")
        assert hasattr(tu.fs.unit.inlet, "pressure")
        assert hasattr(tu.fs.unit.inlet, "alkalinity")

        assert hasattr(tu.fs.unit, "underflow")
        assert len(tu.fs.unit.underflow.vars) == 5
        assert hasattr(tu.fs.unit.underflow, "flow_vol")
        assert hasattr(tu.fs.unit.underflow, "conc_mass_comp")
        assert hasattr(tu.fs.unit.underflow, "temperature")
        assert hasattr(tu.fs.unit.underflow, "pressure")
        assert hasattr(tu.fs.unit.underflow, "alkalinity")

        assert hasattr(tu.fs.unit, "overflow")
        assert len(tu.fs.unit.overflow.vars) == 5
        assert hasattr(tu.fs.unit.overflow, "flow_vol")
        assert hasattr(tu.fs.unit.overflow, "conc_mass_comp")
        assert hasattr(tu.fs.unit.overflow, "temperature")
        assert hasattr(tu.fs.unit.overflow, "pressure")
        assert hasattr(tu.fs.unit.overflow, "alkalinity")

        assert number_variables(tu) == 87
        assert number_total_constraints(tu) == 63
        assert number_unused_variables(tu) == 6

    @pytest.mark.unit
    def test_dof(self, tu):
        assert degrees_of_freedom(tu) == 0

    @pytest.mark.unit
    def test_units(self, tu):
        assert_units_consistent(tu)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, tu):

        iscale.calculate_scaling_factors(tu)
        initialization_tester(tu)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, tu):
        solver = get_solver()
        results = solver.solve(tu)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, tu):
        assert pytest.approx(101325.0, rel=1e-3) == value(
            tu.fs.unit.overflow.pressure[0]
        )
        assert pytest.approx(308.15, rel=1e-3) == value(
            tu.fs.unit.overflow.temperature[0]
        )
        assert pytest.approx(0.003115, rel=1e-3) == value(
            tu.fs.unit.overflow.flow_vol[0]
        )
        assert pytest.approx(0.02806, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "S_I"]
        )
        assert pytest.approx(0.000673, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "S_S"]
        )
        assert pytest.approx(0.06768, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "X_I"]
        )
        assert pytest.approx(0.001409, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "X_S"]
        )
        assert pytest.approx(0.04286, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "X_P"]
        )
        assert pytest.approx(0.099046, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "X_BH"]
        )
        assert pytest.approx(0.007414, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "X_BA"]
        )
        assert pytest.approx(0.001374, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "S_O"]
        )
        assert pytest.approx(0.009194, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "S_NO"]
        )
        assert pytest.approx(0.0001584, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "S_NH"]
        )
        assert pytest.approx(0.0005594, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "S_ND"]
        )
        assert pytest.approx(0.0001056, rel=1e-3) == value(
            tu.fs.unit.overflow.conc_mass_comp[0, "X_ND"]
        )
        assert pytest.approx(0.004564, rel=1e-3) == value(
            tu.fs.unit.overflow.alkalinity[0]
        )
        assert pytest.approx(3.82, rel=1e-3) == value(tu.fs.unit.height)
        assert pytest.approx(300, rel=1e-3) == value(tu.fs.unit.volume[0])
        assert pytest.approx(78.54, rel=1e-3) == value(tu.fs.unit.surface_area)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, tu):
        assert (
            abs(
                value(
                    tu.fs.unit.inlet.flow_vol[0] * tu.fs.props.dens_mass
                    - tu.fs.unit.overflow.flow_vol[0] * tu.fs.props.dens_mass
                    - tu.fs.unit.underflow.flow_vol[0] * tu.fs.props.dens_mass
                )
            )
            <= 1e-6
        )
        for i in tu.fs.props.solute_set:
            assert (
                abs(
                    value(
                        tu.fs.unit.inlet.flow_vol[0]
                        * tu.fs.unit.inlet.conc_mass_comp[0, i]
                        - tu.fs.unit.overflow.flow_vol[0]
                        * tu.fs.unit.overflow.conc_mass_comp[0, i]
                        - tu.fs.unit.underflow.flow_vol[0]
                        * tu.fs.unit.underflow.conc_mass_comp[0, i]
                    )
                )
                <= 1e-6
            )

    @pytest.mark.unit
    def test_report(self, tu):
        tu.fs.unit.report()


class TestThickASM2d(object):
    @pytest.fixture(scope="class")
    def tu_asm2d(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ASM2dParameterBlock()

        m.fs.unit = Thickener(
            property_package=m.fs.props,
            activated_sludge_model=ActivatedSludgeModelType.ASM2D,
        )

        # NOTE: Concentrations of exactly 0 result in singularities, use EPS instead
        EPS = 1e-8

        m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O2"].fix(7.9707 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_N2"].fix(29.0603 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH4"].fix(8.0209 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO3"].fix(6.6395 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_PO4"].fix(7.8953 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_F"].fix(0.4748 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_A"].fix(0.0336 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1695.7695 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(68.2975 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_H"].fix(1855.5067 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(
            214.5319 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(63.5316 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(2.7381 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_AUT"].fix(
            118.3582 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_MeOH"].fix(EPS * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_MeP"].fix(EPS * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_TSS"].fix(
            3525.429 * units.mg / units.liter
        )
        m.fs.unit.inlet.alkalinity[0].fix(4.6663 * units.mmol / units.liter)

        m.fs.unit.hydraulic_retention_time.fix()
        m.fs.unit.diameter.fix()

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, tu_asm2d):

        assert hasattr(tu_asm2d.fs.unit, "inlet")
        assert len(tu_asm2d.fs.unit.inlet.vars) == 5
        assert hasattr(tu_asm2d.fs.unit.inlet, "flow_vol")
        assert hasattr(tu_asm2d.fs.unit.inlet, "conc_mass_comp")
        assert hasattr(tu_asm2d.fs.unit.inlet, "temperature")
        assert hasattr(tu_asm2d.fs.unit.inlet, "pressure")
        assert hasattr(tu_asm2d.fs.unit.inlet, "alkalinity")

        assert hasattr(tu_asm2d.fs.unit, "underflow")
        assert len(tu_asm2d.fs.unit.underflow.vars) == 5
        assert hasattr(tu_asm2d.fs.unit.underflow, "flow_vol")
        assert hasattr(tu_asm2d.fs.unit.underflow, "conc_mass_comp")
        assert hasattr(tu_asm2d.fs.unit.underflow, "temperature")
        assert hasattr(tu_asm2d.fs.unit.underflow, "pressure")
        assert hasattr(tu_asm2d.fs.unit.underflow, "alkalinity")

        assert hasattr(tu_asm2d.fs.unit, "overflow")
        assert len(tu_asm2d.fs.unit.overflow.vars) == 5
        assert hasattr(tu_asm2d.fs.unit.overflow, "flow_vol")
        assert hasattr(tu_asm2d.fs.unit.overflow, "conc_mass_comp")
        assert hasattr(tu_asm2d.fs.unit.overflow, "temperature")
        assert hasattr(tu_asm2d.fs.unit.overflow, "pressure")
        assert hasattr(tu_asm2d.fs.unit.overflow, "alkalinity")

        assert number_variables(tu_asm2d) == 111
        assert number_total_constraints(tu_asm2d) == 87
        assert number_unused_variables(tu_asm2d) == 0

    @pytest.mark.unit
    def test_dof(self, tu_asm2d):
        assert degrees_of_freedom(tu_asm2d) == 0

    @pytest.mark.unit
    def test_units(self, tu_asm2d):
        assert_units_consistent(tu_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, tu_asm2d):

        iscale.calculate_scaling_factors(tu_asm2d)
        initialization_tester(tu_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, tu_asm2d):
        solver = get_solver()
        results = solver.solve(tu_asm2d)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, tu_asm2d):
        assert pytest.approx(101325.0, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.pressure[0]
        )
        assert pytest.approx(308.15, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.temperature[0]
        )
        assert pytest.approx(0.00330, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.flow_vol[0]
        )
        assert pytest.approx(3.36066764339686e-05, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_A"]
        )
        assert pytest.approx(0.0004748063808766291, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_F"]
        )
        assert pytest.approx(0.03, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_I"]
        )
        assert pytest.approx(0.029060299999999997, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_N2"]
        )
        assert pytest.approx(0.00802090132578769, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_NH4"]
        )
        tu_asm2d.fs.unit.overflow.conc_mass_comp.pprint()
        assert pytest.approx(0.006639502251179597, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_NO3"]
        )
        assert pytest.approx(0.007970701359416384, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_O2"]
        )
        assert pytest.approx(0.007895301409926407, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_PO4"]
        )
        assert pytest.approx(0.0024900686245326892, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_AUT"]
        )
        assert pytest.approx(0.03903683632802943, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_H"]
        )
        assert pytest.approx(0.03567622602578816, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_I"]
        )
        assert pytest.approx(0, abs=1e-6) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_MeOH"]
        )
        assert pytest.approx(0, abs=1e-6) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_MeP"]
        )
        assert pytest.approx(0.004513405145476244, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_PAO"]
        )
        assert pytest.approx(5.761182100796471e-05, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_PHA"]
        )
        assert pytest.approx(0.001336607105435848, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_PP"]
        )
        assert pytest.approx(0.0014368738054293484, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_S"]
        )
        assert pytest.approx(0.0741692793990388, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_TSS"]
        )
        assert pytest.approx(0.004666303573014916, rel=1e-3) == value(
            tu_asm2d.fs.unit.overflow.alkalinity[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, tu_asm2d):
        assert (
            abs(
                value(
                    tu_asm2d.fs.unit.inlet.flow_vol[0] * tu_asm2d.fs.props.dens_mass
                    - tu_asm2d.fs.unit.overflow.flow_vol[0]
                    * tu_asm2d.fs.props.dens_mass
                    - tu_asm2d.fs.unit.underflow.flow_vol[0]
                    * tu_asm2d.fs.props.dens_mass
                )
            )
            <= 1e-6
        )
        for i in tu_asm2d.fs.props.solute_set:
            assert (
                abs(
                    value(
                        tu_asm2d.fs.unit.inlet.flow_vol[0]
                        * tu_asm2d.fs.unit.inlet.conc_mass_comp[0, i]
                        - tu_asm2d.fs.unit.overflow.flow_vol[0]
                        * tu_asm2d.fs.unit.overflow.conc_mass_comp[0, i]
                        - tu_asm2d.fs.unit.underflow.flow_vol[0]
                        * tu_asm2d.fs.unit.underflow.conc_mass_comp[0, i]
                    )
                )
                <= 1e-6
            )

    @pytest.mark.unit
    def test_report(self, tu_asm2d):
        tu_asm2d.fs.unit.report()


class TestThickModifiedASM2d(object):
    @pytest.fixture(scope="class")
    def tu_mod_asm2d(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ModifiedASM2dParameterBlock()

        m.fs.unit = Thickener(
            property_package=m.fs.props,
            activated_sludge_model=ActivatedSludgeModelType.modified_ASM2D,
        )

        # NOTE: Concentrations of exactly 0 result in singularities, use EPS instead
        EPS = 1e-8

        m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O2"].fix(7.9707 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_N2"].fix(29.0603 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH4"].fix(8.0209 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO3"].fix(6.6395 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_PO4"].fix(7.8953 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_F"].fix(0.4748 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_A"].fix(0.0336 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_K"].fix(7 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_Mg"].fix(6 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_IC"].fix(10 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1695.7695 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(68.2975 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_H"].fix(1855.5067 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(
            214.5319 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(63.5316 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(2.7381 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_AUT"].fix(
            118.3582 * units.mg / units.liter
        )

        m.fs.unit.hydraulic_retention_time.fix()
        m.fs.unit.diameter.fix()

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, tu_mod_asm2d):

        assert hasattr(tu_mod_asm2d.fs.unit, "inlet")
        assert len(tu_mod_asm2d.fs.unit.inlet.vars) == 4
        assert hasattr(tu_mod_asm2d.fs.unit.inlet, "flow_vol")
        assert hasattr(tu_mod_asm2d.fs.unit.inlet, "conc_mass_comp")
        assert hasattr(tu_mod_asm2d.fs.unit.inlet, "temperature")
        assert hasattr(tu_mod_asm2d.fs.unit.inlet, "pressure")

        assert hasattr(tu_mod_asm2d.fs.unit, "underflow")
        assert len(tu_mod_asm2d.fs.unit.underflow.vars) == 4
        assert hasattr(tu_mod_asm2d.fs.unit.underflow, "flow_vol")
        assert hasattr(tu_mod_asm2d.fs.unit.underflow, "conc_mass_comp")
        assert hasattr(tu_mod_asm2d.fs.unit.underflow, "temperature")
        assert hasattr(tu_mod_asm2d.fs.unit.underflow, "pressure")

        assert hasattr(tu_mod_asm2d.fs.unit, "overflow")
        assert len(tu_mod_asm2d.fs.unit.overflow.vars) == 4
        assert hasattr(tu_mod_asm2d.fs.unit.overflow, "flow_vol")
        assert hasattr(tu_mod_asm2d.fs.unit.overflow, "conc_mass_comp")
        assert hasattr(tu_mod_asm2d.fs.unit.overflow, "temperature")
        assert hasattr(tu_mod_asm2d.fs.unit.overflow, "pressure")

        assert number_variables(tu_mod_asm2d) == 115
        assert number_total_constraints(tu_mod_asm2d) == 86
        assert number_unused_variables(tu_mod_asm2d) == 0

    @pytest.mark.unit
    def test_dof(self, tu_mod_asm2d):
        assert degrees_of_freedom(tu_mod_asm2d) == 0

    @pytest.mark.unit
    def test_units(self, tu_mod_asm2d):
        assert_units_consistent(tu_mod_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, tu_mod_asm2d):

        iscale.calculate_scaling_factors(tu_mod_asm2d)
        initialization_tester(tu_mod_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, tu_mod_asm2d):
        solver = get_solver()
        results = solver.solve(tu_mod_asm2d)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, tu_mod_asm2d):
        assert pytest.approx(101325.0, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.pressure[0]
        )
        assert pytest.approx(308.15, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.temperature[0]
        )
        assert pytest.approx(0.0033139, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.flow_vol[0]
        )
        assert pytest.approx(3.36066764339686e-05, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_A"]
        )
        assert pytest.approx(0.0004748063808766291, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_F"]
        )
        assert pytest.approx(0.03, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_I"]
        )
        assert pytest.approx(0.029060299999999997, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_N2"]
        )
        assert pytest.approx(0.00802090132578769, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_NH4"]
        )
        tu_mod_asm2d.fs.unit.overflow.conc_mass_comp.pprint()
        assert pytest.approx(0.006639502251179597, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_NO3"]
        )
        assert pytest.approx(0.007970701359416384, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_O2"]
        )
        assert pytest.approx(0.007895301409926407, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_PO4"]
        )
        assert pytest.approx(0.006999, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_K"]
        )
        assert pytest.approx(0.00599999, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_Mg"]
        )
        assert pytest.approx(0.0099999999, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "S_IC"]
        )
        assert pytest.approx(0.0024802, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_AUT"]
        )
        assert pytest.approx(0.0388828, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_H"]
        )
        assert pytest.approx(0.0355354585, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_I"]
        )
        assert pytest.approx(0.00449559, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_PAO"]
        )
        assert pytest.approx(5.737786825381196e-05, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_PHA"]
        )
        assert pytest.approx(0.001331327, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_PP"]
        )
        assert pytest.approx(0.0014311986, rel=1e-3) == value(
            tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, "X_S"]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, tu_mod_asm2d):
        assert (
            abs(
                value(
                    tu_mod_asm2d.fs.unit.inlet.flow_vol[0]
                    * tu_mod_asm2d.fs.props.dens_mass
                    - tu_mod_asm2d.fs.unit.overflow.flow_vol[0]
                    * tu_mod_asm2d.fs.props.dens_mass
                    - tu_mod_asm2d.fs.unit.underflow.flow_vol[0]
                    * tu_mod_asm2d.fs.props.dens_mass
                )
            )
            <= 1e-6
        )
        for i in tu_mod_asm2d.fs.props.solute_set:
            assert (
                abs(
                    value(
                        tu_mod_asm2d.fs.unit.inlet.flow_vol[0]
                        * tu_mod_asm2d.fs.unit.inlet.conc_mass_comp[0, i]
                        - tu_mod_asm2d.fs.unit.overflow.flow_vol[0]
                        * tu_mod_asm2d.fs.unit.overflow.conc_mass_comp[0, i]
                        - tu_mod_asm2d.fs.unit.underflow.flow_vol[0]
                        * tu_mod_asm2d.fs.unit.underflow.conc_mass_comp[0, i]
                    )
                )
                <= 1e-6
            )

    @pytest.mark.unit
    def test_report(self, tu_mod_asm2d):
        tu_mod_asm2d.fs.unit.report()


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_costing():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ModifiedASM2dParameterBlock()

    m.fs.unit = Thickener(
        property_package=m.fs.props,
        activated_sludge_model=ActivatedSludgeModelType.modified_ASM2D,
    )

    # NOTE: Concentrations of exactly 0 result in singularities, use EPS instead
    EPS = 1e-8

    m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O2"].fix(7.9707 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_N2"].fix(29.0603 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH4"].fix(8.0209 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO3"].fix(6.6395 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_PO4"].fix(7.8953 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_F"].fix(0.4748 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_A"].fix(0.0336 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_K"].fix(7 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_Mg"].fix(6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_IC"].fix(10 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1695.7695 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(68.2975 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_H"].fix(1855.5067 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(214.5319 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(63.5316 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(2.7381 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_AUT"].fix(118.3582 * units.mg / units.liter)

    m.fs.unit.hydraulic_retention_time.fix()
    m.fs.unit.diameter.fix()

    m.fs.costing = WaterTAPCosting()

    m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    m.fs.costing.cost_process()

    assert degrees_of_freedom(m) == 0

    results = solver.solve(m)

    assert_optimal_termination(results)
    assert_units_consistent(m)
    assert hasattr(m.fs.costing, "thickener")
    assert value(m.fs.costing.thickener.capital_a_parameter) == 4729.8
    assert value(m.fs.costing.thickener.capital_b_parameter) == 37068

    # Check solutions
    assert pytest.approx(220675.79 * 2, rel=1e-5) == value(
        m.fs.unit.costing.capital_cost
    )
    assert pytest.approx(220675.79, rel=1e-5) == value(
        units.convert(
            (4729.8 * value(units.convert(10 * units.m, to_units=units.feet)) + 37068)
            * units.USD_2007,
            to_units=m.fs.costing.base_currency,
        )
    )
    assert pytest.approx(12.5 * 0.01255, rel=1e-5) == value(
        m.fs.unit.electricity_consumption[0]
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""

"""

# Some more information about this module
__author__ = "Alejandro Garciadiego, Andrew Lee"

import pyomo.environ as pyo
from pyomo.network import Arc, SequentialDecomposition

from idaes.core import FlowsheetBlock
from idaes.models.unit_models import (
    CSTR,
    Feed,
    Mixer,
    Separator,
    PressureChanger,
    Product,
)
from idaes.models.unit_models.separator import SplittingType
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.tables import (
    create_stream_table_dataframe,
    stream_table_dataframe_to_string,
)

from watertap.unit_models.cstr_injection import CSTR_Injection
from watertap.property_models.activated_sludge.asm2d_properties import (
    ASM2dParameterBlock,
)
from watertap.property_models.activated_sludge.asm2d_reactions import (
    ASM2dReactionParameterBlock,
)

from watertap.core.util.initialization import check_solve

# Set up logger
_log = idaeslog.getLogger(__name__)


def build_flowsheet():
    m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM2dParameterBlock()
    m.fs.rxn_props = ASM2dReactionParameterBlock(property_package=m.fs.props)

    # Feed water stream
    m.fs.FeedWater = Feed(property_package=m.fs.props)
    # Mixer for feed water and recycled sludge
    m.fs.MX1 = Mixer(property_package=m.fs.props, inlet_list=["feed_water", "recycle"])
    # First reactor (anoxic) - standard CSTR
    m.fs.R1 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # Second reactor (anoxic) - standard CSTR
    m.fs.R2 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # Third reactor (aerobic) - CSTR with injection
    m.fs.R3 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    # Fourth reactor (aerobic) - CSTR with injection
    m.fs.R4 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    # Fifth reactor (aerobic) - CSTR with injection
    m.fs.R5 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    m.fs.SP5 = Separator(
        property_package=m.fs.props, outlet_list=["underflow", "overflow"]
    )
    # Clarifier
    # TODO: Replace with more detailed model when available
    m.fs.CL1 = Separator(
        property_package=m.fs.props,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )
    # Sludge purge splitter
    m.fs.SP6 = Separator(
        property_package=m.fs.props,
        outlet_list=["recycle", "waste"],
        split_basis=SplittingType.totalFlow,
    )
    # Mixing sludge recycle and R5 underflow
    m.fs.MX6 = Mixer(property_package=m.fs.props, inlet_list=["clarifier", "reactor"])
    # Product Blocks
    m.fs.Treated = Product(property_package=m.fs.props)
    m.fs.Sludge = Product(property_package=m.fs.props)
    # Recycle pressure changer - use a simple isothermal unit for now
    m.fs.P1 = PressureChanger(property_package=m.fs.props)

    # Link units
    m.fs.stream1 = Arc(source=m.fs.FeedWater.outlet, destination=m.fs.MX1.feed_water)
    m.fs.stream2 = Arc(source=m.fs.MX1.outlet, destination=m.fs.R1.inlet)
    m.fs.stream3 = Arc(source=m.fs.R1.outlet, destination=m.fs.R2.inlet)
    m.fs.stream4 = Arc(source=m.fs.R2.outlet, destination=m.fs.R3.inlet)
    m.fs.stream5 = Arc(source=m.fs.R3.outlet, destination=m.fs.R4.inlet)
    m.fs.stream6 = Arc(source=m.fs.R4.outlet, destination=m.fs.R5.inlet)
    m.fs.stream7 = Arc(source=m.fs.R5.outlet, destination=m.fs.SP5.inlet)
    m.fs.stream8 = Arc(source=m.fs.SP5.overflow, destination=m.fs.CL1.inlet)
    m.fs.stream9 = Arc(source=m.fs.SP5.underflow, destination=m.fs.MX6.reactor)
    m.fs.stream10 = Arc(source=m.fs.CL1.effluent, destination=m.fs.Treated.inlet)
    m.fs.stream11 = Arc(source=m.fs.CL1.underflow, destination=m.fs.SP6.inlet)
    m.fs.stream102 = Arc(source=m.fs.SP6.waste, destination=m.fs.Sludge.inlet)
    m.fs.stream13 = Arc(source=m.fs.SP6.recycle, destination=m.fs.MX6.clarifier)
    m.fs.stream14 = Arc(source=m.fs.MX6.outlet, destination=m.fs.P1.inlet)
    m.fs.stream15 = Arc(source=m.fs.P1.outlet, destination=m.fs.MX1.recycle)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Oxygen concentration in reactors 3 and 4 is governed by mass transfer
    # Add additional parameter and constraints
    m.fs.R3.KLa = pyo.Var(
        initialize=3.4,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.R4.KLa = pyo.Var(
        initialize=1.8,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.S_O_eq = pyo.Param(
        default=8e-3,
        units=pyo.units.kg / pyo.units.m**3,
        mutable=True,
        doc="Dissolved oxygen concentration at equilibrium",
    )

    m.fs.R3.Constraint(m.fs.time, doc="Mass transfer constraint for R3")

    def mass_transfer_R3(self, t):
        return pyo.units.convert(
            m.fs.R3.injection[t, "Liq", "S_O2"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R3.KLa
            * m.fs.R3.volume[t]
            * (m.fs.S_O_eq - m.fs.R3.outlet.conc_mass_comp[t, "S_O2"])
        )

    m.fs.R4.Constraint(m.fs.time, doc="Mass transfer constraint for R4")

    def mass_transfer_R4(self, t):
        return pyo.units.convert(
            m.fs.R4.injection[t, "Liq", "S_O2"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R4.KLa
            * m.fs.R4.volume[t]
            * (m.fs.S_O_eq - m.fs.R4.outlet.conc_mass_comp[t, "S_O2"])
        )

    # Feed Water Conditions
    m.fs.FeedWater.flow_vol.fix(18446 * pyo.units.m**3 / pyo.units.day)
    m.fs.FeedWater.temperature.fix(298.15 * pyo.units.K)
    m.fs.FeedWater.pressure.fix(1 * pyo.units.atm)
    m.fs.FeedWater.conc_mass_comp[0, "S_O2"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_F"].fix(30 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_A"].fix(20 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NH4"].fix(16 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NO3"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_PO4"].fix(3.6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_I"].fix(30 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_N2"].fix(15 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_I"].fix(25 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_S"].fix(125 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_H"].fix(30 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_PAO"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_PP"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_PHA"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_AUT"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_MeOH"].fix(
        1e-6 * pyo.units.g / pyo.units.m**3
    )
    m.fs.FeedWater.conc_mass_comp[0, "X_MeP"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_TSS"].fix(180 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.alkalinity.fix(7 * pyo.units.mol / pyo.units.m**3)

    # Reactor sizing
    m.fs.R1.volume.fix(1000 * pyo.units.m**3)
    m.fs.R2.volume.fix(1000 * pyo.units.m**3)
    m.fs.R3.volume.fix(1333 * pyo.units.m**3)
    m.fs.R4.volume.fix(1333 * pyo.units.m**3)
    m.fs.R5.volume.fix(1333 * pyo.units.m**3)

    # Injection rates to Reactions 3, 4 and 5
    for j in m.fs.props.component_list:
        if j != "S_O2":
            # All components except S_O have no injection
            m.fs.R3.injection[:, :, j].fix(0)
            m.fs.R4.injection[:, :, j].fix(0)
            m.fs.R5.injection[:, :, j].fix(0)
    # Then set injections rates for O2
    m.fs.R3.outlet.conc_mass_comp[:, "S_O2"].fix(2.72e-3)
    m.fs.R4.outlet.conc_mass_comp[:, "S_O2"].fix(2.43e-3)
    m.fs.R5.outlet.conc_mass_comp[:, "S_O2"].fix(4.49e-4)

    # Set fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP5.split_fraction[:, "underflow"].fix(0.1)

    # Clarifier
    # TODO: Update once more detailed model available
    m.fs.CL1.split_fraction[0, "effluent", "H2O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_A"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_F"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_I"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_N2"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NH4"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NO3"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_O2"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_PO4"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_ALK"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_AUT"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_H"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_I"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_MeOH"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_MeP"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_PAO"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_PHA"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_PP"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_S"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_TSS"].fix(0.00187)

    # Sludge purge separator
    m.fs.SP6.split_fraction[:, "recycle"].fix(0.65)

    # Outlet pressure from recycle pump
    m.fs.P1.outlet.pressure.fix(101325)

    # Check degrees of freedom
    print(degrees_of_freedom(m))
    assert degrees_of_freedom(m) == 0

    def scale_variables(m):
        for var in m.fs.component_data_objects(pyo.Var, descend_into=True):
            if "flow_vol" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "temperature" in var.name:
                iscale.set_scaling_factor(var, 1e-1)
            if "pressure" in var.name:
                iscale.set_scaling_factor(var, 1e-4)
            if "enth_mol" in var.name:
                iscale.set_scaling_factor(var, 1e-4)
            if "conc_mass_comp[S_O2]" in var.name:
                iscale.set_scaling_factor(var, 1e5)
            if "conc_mass_comp[S_F]" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "conc_mass_comp[S_A]" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "conc_mass_comp[S_NH4]" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "conc_mass_comp[S_NO3]" in var.name:
                iscale.set_scaling_factor(var, 1e5)
            if "conc_mass_comp[S_PO4]" in var.name:
                iscale.set_scaling_factor(var, 1e3)
            if "conc_mass_comp[S_I]" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "conc_mass_comp[S_N2]" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "conc_mass_comp[X_I]" in var.name:
                iscale.set_scaling_factor(var, 1e0)
            if "conc_mass_comp[X_S]" in var.name:
                iscale.set_scaling_factor(var, 1e1)
            if "conc_mass_comp[X_H]" in var.name:
                iscale.set_scaling_factor(var, 1e0)
            if "conc_mass_comp[X_PAO]" in var.name:
                iscale.set_scaling_factor(var, 1e1)
            if "conc_mass_comp[X_PP]" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "conc_mass_comp[X_PHA]" in var.name:
                iscale.set_scaling_factor(var, 1e2)
            if "conc_mass_comp[X_AUT]" in var.name:
                iscale.set_scaling_factor(var, 1e5)
            if "conc_mass_comp[X_MeOH]" in var.name:
                iscale.set_scaling_factor(var, 1e5)
            if "conc_mass_comp[X_MeP]" in var.name:
                iscale.set_scaling_factor(var, 1e5)
            if "conc_mass_comp[X_TSS]" in var.name:
                iscale.set_scaling_factor(var, 1e0)

    scale_variables(m)
    iscale.calculate_scaling_factors(m.fs)

    # Initialize flowsheet
    # Apply sequential decomposition - 1 iteration should suffice
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 0

    G = seq.create_graph(m)

    # # Uncomment this code to see tear set and initialization order
    # heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
    # order = seq.calculation_order(G)
    # for o in heuristic_tear_set:
    #     print(o.name)
    # for o in order:
    #     print(o[0].name)

    # Initial guesses for flow into first reactor
    tear_guesses = {
        "flow_vol": {0: 0.31},
        "conc_mass_comp": {
            (0, "S_O2"): 0.14,
            (0, "S_F"): 0.021,
            (0, "S_A"): 0.014,
            (0, "S_NH4"): 0.016,
            (0, "S_NO3"): 1e-6,
            (0, "S_PO4"): 0.0035,
            (0, "S_I"): 0.030,
            (0, "S_N2"): 0.015,
            (0, "X_I"): 0.050,
            (0, "X_S"): 0.138,
            (0, "X_H"): 0.014,
            (0, "X_PAO"): 1e-6,
            (0, "X_PP"): 1e-6,
            (0, "X_PHA"): 1e-6,
            (0, "X_AUT"): 1e-6,
            (0, "X_MeOH"): 1e-6,
            (0, "X_MeP"): 1e-6,
            (0, "X_TSS"): 0.366,
        },
        "alkalinity": {0: 7.2e-3},
        "temperature": {0: 298.15},
        "pressure": {0: 101325},
    }

    # Pass the tear_guess to the SD tool
    seq.set_guesses_for(m.fs.R1.inlet, tear_guesses)

    def function(unit):
        unit.initialize(outlvl=idaeslog.INFO, optarg={"bound_push": 1e-8})

    seq.run(m, function)

    solver = get_solver()
    results = solver.solve(m, tee=False)
    check_solve(results, logger=_log, fail_flag=True)

    return m, results


if __name__ == "__main__":
    # This method builds and runs a steady state activated sludge
    # flowsheet.
    m, results = build_flowsheet()

    stream_table = create_stream_table_dataframe(
        {
            "Feed": m.fs.FeedWater.outlet,
            "Mix": m.fs.R1.inlet,
            "R1": m.fs.R1.outlet,
            "R2": m.fs.R2.outlet,
            "R3": m.fs.R3.outlet,
            "R4": m.fs.R4.outlet,
            "R5": m.fs.R5.outlet,
        },
        time_point=0,
    )
    print(stream_table_dataframe_to_string(stream_table))



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of constructed wetlands
for wastewater resource recovery flowsheets.
"""

import pyomo.environ as pyo
from idaes.core import declare_process_block_class
from watertap.core import build_siso, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("ConstructedWetlandsZO")
class ConstructedWetlandsZOData(ZeroOrderBaseData):
    """
    Zero-Order model for constructed wetlands.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "constructed_wetlands"

        build_siso(self)

    @property
    def default_costing_method(self):
        return self.cost_constructed_wetlands

    @staticmethod
    def cost_constructed_wetlands(blk):
        """
        Method for costing constructed wetlands.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )
        # Get costing parameter sub-block for this technology
        unit_capex = blk.unit_model._get_tech_parameters(
            blk, parameter_dict, blk.unit_model.config.process_subtype, ["unit_capex"]
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        capex_expr = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol * unit_capex,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * capex_expr
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    value,
    Constraint,
    Objective,
    Param,
    TransformationFactory,
    assert_optimal_termination,
)
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import solve_indexed_blocks, propagate_state
from idaes.models.unit_models import Mixer, Separator, Product, Feed
from idaes.models.unit_models.mixer import MomentumMixingType
from idaes.core import UnitModelCostingBlock
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.misc import StrEnum

import watertap.property_models.NaCl_prop_pack as props
from watertap.unit_models.reverse_osmosis_0D import (
    ReverseOsmosis0D,
    ConcentrationPolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
)
from watertap.unit_models.pressure_exchanger import PressureExchanger
from watertap.unit_models.pressure_changer import Pump, EnergyRecoveryDevice
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.costing import WaterTAPCosting


class ERDtype(StrEnum):
    pressure_exchanger = "pressure_exchanger"
    pump_as_turbine = "pump_as_turbine"


def erd_type_not_found(erd_type):
    raise NotImplementedError(
        "erd_type was {}, but can only "
        "be pressure_exchanger or pump_as_turbine"
        "".format(erd_type.value)
    )


def main(erd_type=ERDtype.pressure_exchanger):
    # set up solver
    solver = get_solver()

    # build, set, and initialize
    m = build(erd_type=erd_type)
    set_operating_conditions(m)
    initialize_system(m, solver=solver)

    # optimize and display
    optimize_set_up(m)
    solve(m, solver=solver)

    print("\n***---Simulation results---***")
    display_system(m)
    display_design(m)
    if erd_type == ERDtype.pressure_exchanger:
        display_state(m)
    else:
        pass

    return m


def build(erd_type=ERDtype.pressure_exchanger):
    # flowsheet set up
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.erd_type = erd_type
    m.fs.properties = props.NaClParameterBlock()
    m.fs.costing = WaterTAPCosting()

    # Control volume flow blocks
    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)

    # --- Main pump ---
    m.fs.P1 = Pump(property_package=m.fs.properties)
    m.fs.P1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # --- Reverse Osmosis Block ---
    m.fs.RO = ReverseOsmosis0D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        pressure_change_type=PressureChangeType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
    )
    m.fs.RO.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # --- ERD blocks ---
    if erd_type == ERDtype.pressure_exchanger:
        m.fs.S1 = Separator(property_package=m.fs.properties, outlet_list=["P1", "PXR"])

        m.fs.PXR = PressureExchanger(property_package=m.fs.properties)
        m.fs.P2 = Pump(property_package=m.fs.properties)
        m.fs.M1 = Mixer(
            property_package=m.fs.properties,
            momentum_mixing_type=MomentumMixingType.equality,
            inlet_list=["P1", "P2"],
        )

        # add costing for PX and recirculation pump
        m.fs.PXR.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.P2.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        # mixer and separator have no associated costing
    elif erd_type == ERDtype.pump_as_turbine:
        # add energy recovery turbine block
        m.fs.ERD = EnergyRecoveryDevice(property_package=m.fs.properties)
        # add costing for ERD config
        m.fs.ERD.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    else:
        erd_type_not_found(erd_type)

    # process costing and add system level metrics
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_specific_electrical_carbon_intensity(
        m.fs.product.properties[0].flow_vol
    )

    # connections
    if erd_type == ERDtype.pressure_exchanger:
        m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.S1.inlet)
        m.fs.s02 = Arc(source=m.fs.S1.P1, destination=m.fs.P1.inlet)
        m.fs.s03 = Arc(source=m.fs.P1.outlet, destination=m.fs.M1.P1)
        m.fs.s04 = Arc(source=m.fs.M1.outlet, destination=m.fs.RO.inlet)
        m.fs.s05 = Arc(source=m.fs.RO.permeate, destination=m.fs.product.inlet)
        m.fs.s06 = Arc(
            source=m.fs.RO.retentate, destination=m.fs.PXR.high_pressure_inlet
        )
        m.fs.s07 = Arc(
            source=m.fs.PXR.high_pressure_outlet, destination=m.fs.disposal.inlet
        )
        m.fs.s08 = Arc(source=m.fs.S1.PXR, destination=m.fs.PXR.low_pressure_inlet)
        m.fs.s09 = Arc(source=m.fs.PXR.low_pressure_outlet, destination=m.fs.P2.inlet)
        m.fs.s10 = Arc(source=m.fs.P2.outlet, destination=m.fs.M1.P2)
    elif erd_type == ERDtype.pump_as_turbine:
        m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.P1.inlet)
        m.fs.s02 = Arc(source=m.fs.P1.outlet, destination=m.fs.RO.inlet)
        m.fs.s03 = Arc(source=m.fs.RO.permeate, destination=m.fs.product.inlet)
        m.fs.s04 = Arc(source=m.fs.RO.retentate, destination=m.fs.ERD.inlet)
        m.fs.s05 = Arc(source=m.fs.ERD.outlet, destination=m.fs.disposal.inlet)
    else:
        # this case should be caught in the previous conditional
        erd_type_not_found(erd_type)

    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    # set default property values
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "NaCl")
    )
    # set unit model values
    iscale.set_scaling_factor(m.fs.P1.control_volume.work, 1e-3)
    iscale.set_scaling_factor(m.fs.RO.area, 1e-2)
    m.fs.feed.properties[0].flow_vol_phase["Liq"]
    m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "NaCl"]
    if erd_type == ERDtype.pressure_exchanger:
        iscale.set_scaling_factor(m.fs.P2.control_volume.work, 1e-3)
        iscale.set_scaling_factor(m.fs.PXR.low_pressure_side.work, 1e-3)
        iscale.set_scaling_factor(m.fs.PXR.high_pressure_side.work, 1e-3)
        # touch properties used in specifying and initializing the model
        m.fs.S1.mixed_state[0].mass_frac_phase_comp
        m.fs.S1.PXR_state[0].flow_vol_phase["Liq"]
    elif erd_type == ERDtype.pump_as_turbine:
        iscale.set_scaling_factor(m.fs.ERD.control_volume.work, 1e-3)
    else:
        erd_type_not_found(erd_type)
    # unused scaling factors needed by IDAES base costing module
    # calculate and propagate scaling factors
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(
    m,
    water_recovery=0.5,
    over_pressure=0.3,
    solver=None,
):

    if solver is None:
        solver = get_solver()
    # ---specifications---
    # feed
    # state variables
    m.fs.feed.properties[0].pressure.fix(101325)  # feed pressure [Pa]
    m.fs.feed.properties[0].temperature.fix(273.15 + 25)  # feed temperature [K]
    # properties (cannot be fixed for initialization routines, must calculate the state variables)
    m.fs.feed.properties.calculate_state(
        var_args={
            ("flow_vol_phase", "Liq"): 1e-3,  # feed volumetric flow rate [m3/s]
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.035,
        },  # feed NaCl mass fraction [-]
        hold_state=True,  # fixes the calculated component mass flow rates
    )

    # pump 1, high pressure pump, 2 degrees of freedom (efficiency and outlet pressure)
    m.fs.P1.efficiency_pump.fix(0.80)  # pump efficiency [-]
    operating_pressure = calculate_operating_pressure(
        feed_state_block=m.fs.feed.properties[0],
        over_pressure=over_pressure,
        water_recovery=water_recovery,
        NaCl_passage=0.01,
        solver=solver,
    )
    m.fs.P1.control_volume.properties_out[0].pressure.fix(operating_pressure)

    # RO unit
    m.fs.RO.A_comp.fix(4.2e-12)  # membrane water permeability coefficient [m/s-Pa]
    m.fs.RO.B_comp.fix(3.5e-8)  # membrane salt permeability coefficient [m/s]
    m.fs.RO.feed_side.channel_height.fix(1e-3)  # channel height in membrane stage [m]
    m.fs.RO.feed_side.spacer_porosity.fix(0.97)  # spacer porosity in membrane stage [-]
    m.fs.RO.permeate.pressure[0].fix(101325)  # atmospheric pressure [Pa]
    m.fs.RO.width.fix(5)  # stage width [m]
    # initialize RO
    m.fs.RO.feed_side.properties_in[0].flow_mass_phase_comp["Liq", "H2O"] = value(
        m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    )
    m.fs.RO.feed_side.properties_in[0].flow_mass_phase_comp["Liq", "NaCl"] = value(
        m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "NaCl"]
    )
    m.fs.RO.feed_side.properties_in[0].temperature = value(
        m.fs.feed.properties[0].temperature
    )
    m.fs.RO.feed_side.properties_in[0].pressure = value(
        m.fs.P1.control_volume.properties_out[0].pressure
    )

    m.fs.RO.area.fix(50)  # guess area for RO initialization

    if m.fs.erd_type == ERDtype.pressure_exchanger:
        # pressure exchanger
        m.fs.PXR.efficiency_pressure_exchanger.fix(
            0.95
        )  # pressure exchanger efficiency [-]
        # pump 2, booster pump, 1 degree of freedom (efficiency, pressure must match high pressure pump)
        m.fs.P2.efficiency_pump.fix(0.80)

    elif m.fs.erd_type == ERDtype.pump_as_turbine:
        # energy recovery turbine - efficiency and outlet pressure
        m.fs.ERD.efficiency_pump.fix(0.95)
        m.fs.ERD.control_volume.properties_out[0].pressure.fix(101325)
    else:
        erd_type_not_found(m.fs.erd_type)

    m.fs.RO.initialize(optarg=solver.options)

    # unfix guessed area, and fix water recovery
    m.fs.RO.area.unfix()

    m.fs.RO.recovery_mass_phase_comp[0, "Liq", "H2O"].fix(water_recovery)

    # check degrees of freedom
    if degrees_of_freedom(m) != 0:
        raise RuntimeError(
            "The set_operating_conditions function resulted in {} "
            "degrees of freedom rather than 0. This error suggests "
            "that too many or not enough variables are fixed for a "
            "simulation.".format(degrees_of_freedom(m))
        )


def calculate_operating_pressure(
    feed_state_block=None,
    over_pressure=0.15,
    water_recovery=0.5,
    NaCl_passage=0.01,
    solver=None,
):
    """
    estimate operating pressure for RO unit model given the following arguments:

    Arguments:
        feed_state_block:   the state block of the RO feed that has the non-pressure state
                            variables initialized to their values (default=None)
        over_pressure:  the amount of operating pressure above the brine osmotic pressure
                        represented as a fraction (default=0.15)
        water_recovery: the mass-based fraction of inlet H2O that becomes permeate
                        (default=0.5)
        NaCl_passage:   the mass-based fraction of inlet NaCl that becomes permeate
                        (default=0.01)
        solver:     solver object to be used (default=None)
    """
    t = ConcreteModel()  # create temporary model
    prop = feed_state_block.config.parameters
    t.brine = prop.build_state_block([0])

    # specify state block
    t.brine[0].flow_mass_phase_comp["Liq", "H2O"].fix(
        value(feed_state_block.flow_mass_phase_comp["Liq", "H2O"])
        * (1 - water_recovery)
    )
    t.brine[0].flow_mass_phase_comp["Liq", "NaCl"].fix(
        value(feed_state_block.flow_mass_phase_comp["Liq", "NaCl"]) * (1 - NaCl_passage)
    )
    t.brine[0].pressure.fix(
        101325
    )  # valid when osmotic pressure is independent of hydraulic pressure
    t.brine[0].temperature.fix(value(feed_state_block.temperature))

    # calculate osmotic pressure
    # since properties are created on demand, we must touch the property to create it
    t.brine[0].pressure_osm_phase
    # solve state block
    results = solve_indexed_blocks(solver, [t.brine])
    assert_optimal_termination(results)

    return value(t.brine[0].pressure_osm_phase["Liq"]) * (1 + over_pressure)


def solve(blk, solver=None, tee=False, check_termination=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    if check_termination:
        assert_optimal_termination(results)
    return results


def initialize_system(m, solver=None):
    if solver is None:
        solver = get_solver()
    optarg = solver.options

    # ---initialize RO---
    m.fs.RO.initialize(optarg=optarg)

    # ---initialize feed block---
    m.fs.feed.initialize(optarg=optarg)

    # --- initialize ERD ---
    if m.fs.erd_type == ERDtype.pressure_exchanger:
        initialize_pressure_exchanger(m, optarg)

    elif m.fs.erd_type == ERDtype.pump_as_turbine:
        initialize_pump_as_turbine(m, optarg)

    else:
        erd_type_not_found(m.fs.erd_type)

    m.fs.costing.initialize()


def initialize_pressure_exchanger(m, optarg):
    # ---initialize splitter and pressure exchanger---
    # pressure exchanger high pressure inlet
    propagate_state(m.fs.s06)  # propagate to PXR high pressure inlet from RO retentate
    m.fs.PXR.high_pressure_side.properties_in.initialize(optarg=optarg)

    # splitter inlet
    propagate_state(m.fs.s01)  # propagate to splitter inlet from feed
    m.fs.S1.mixed_state.initialize(
        optarg=optarg
    )  # initialize inlet state block to solve for mass fraction
    # splitter outlet to PXR, enforce same volumetric flow as PXR high pressure inlet
    m.fs.S1.PXR_state.calculate_state(
        var_args={
            (
                "flow_vol_phase",
                "Liq",
            ): value(  # same volumetric flow rate as PXR high pressure inlet
                m.fs.PXR.high_pressure_side.properties_in[0].flow_vol_phase["Liq"]
            ),
            ("mass_frac_phase_comp", ("Liq", "NaCl")): value(
                m.fs.S1.mixed_state[0].mass_frac_phase_comp["Liq", "NaCl"]
            ),  # same as splitter inlet
            ("pressure", None): value(
                m.fs.S1.mixed_state[0].pressure
            ),  # same as splitter inlet
            ("temperature", None): value(m.fs.S1.mixed_state[0].temperature),
        },  # same as splitter inlet
    )
    # splitter initialization
    m.fs.S1.PXR_state[0].flow_mass_phase_comp[
        "Liq", "NaCl"
    ].fix()  # fix the single degree of freedom for unit
    m.fs.S1.initialize(optarg=optarg)
    m.fs.S1.PXR_state[0].flow_mass_phase_comp[
        "Liq", "NaCl"
    ].unfix()  # unfix for flowsheet simulation and optimization

    # pressure exchanger low pressure inlet
    propagate_state(m.fs.s08)

    # pressure exchanger initialization
    m.fs.PXR.initialize(optarg=optarg)

    # ---initialize pump 1---
    propagate_state(m.fs.s02)
    m.fs.P1.initialize(optarg=optarg)

    # ---initialize pump 2---
    propagate_state(m.fs.s09)
    m.fs.P2.control_volume.properties_out[0].pressure.fix(
        value(m.fs.P2.control_volume.properties_out[0].pressure)
    )
    m.fs.P2.initialize(optarg=optarg)
    m.fs.P2.control_volume.properties_out[0].pressure.unfix()

    # ---initialize mixer---
    propagate_state(m.fs.s03)
    propagate_state(m.fs.s10)
    m.fs.M1.initialize(optarg=optarg, outlvl=idaeslog.INFO)


def initialize_pump_as_turbine(m, optarg):
    propagate_state(m.fs.s05)
    m.fs.ERD.initialize(optarg=optarg)
    propagate_state(m.fs.s01)
    m.fs.P1.initialize(optarg=optarg)
    propagate_state(m.fs.s02)


def optimize_set_up(m):
    # add objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # unfix decision variables and add bounds
    # pump 1 and pump 2
    m.fs.P1.control_volume.properties_out[0].pressure.unfix()
    m.fs.P1.control_volume.properties_out[0].pressure.setlb(10e5)
    m.fs.P1.control_volume.properties_out[0].pressure.setub(80e5)
    m.fs.P1.deltaP.setlb(0)
    if m.fs.erd_type == ERDtype.pressure_exchanger:
        m.fs.P2.control_volume.properties_out[0].pressure.setlb(10e5)
        m.fs.P2.control_volume.properties_out[0].pressure.setub(80e5)
        m.fs.P2.deltaP.setlb(0)
    else:
        # no additional optimization needed for pump_as_turbine configuration
        pass

    # RO
    m.fs.RO.area.unfix()
    m.fs.RO.area.setlb(1)
    m.fs.RO.area.setub(150)

    # additional specifications
    m.fs.product_salinity = Param(
        initialize=500e-6, mutable=True
    )  # product NaCl mass fraction [-]
    m.fs.minimum_water_flux = Param(
        initialize=1.0 / 3600.0, mutable=True
    )  # minimum water flux [kg/m2-s]

    # additional constraints
    m.fs.eq_product_quality = Constraint(
        expr=m.fs.product.properties[0].mass_frac_phase_comp["Liq", "NaCl"]
        <= m.fs.product_salinity
    )
    iscale.constraint_scaling_transform(
        m.fs.eq_product_quality, 1e3
    )  # scaling constraint
    m.fs.eq_minimum_water_flux = Constraint(
        expr=m.fs.RO.flux_mass_phase_comp[0, 1, "Liq", "H2O"] >= m.fs.minimum_water_flux
    )

    # ---checking model---
    assert_degrees_of_freedom(m, 1)


def optimize(m, solver=None, check_termination=True):
    # --solve---
    return solve(m, solver=solver, check_termination=check_termination)


def display_system(m):
    print("---system metrics---")
    feed_flow_mass = sum(
        m.fs.feed.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    feed_mass_frac_NaCl = (
        m.fs.feed.flow_mass_phase_comp[0, "Liq", "NaCl"].value / feed_flow_mass
    )
    print("Feed: %.2f kg/s, %.0f ppm" % (feed_flow_mass, feed_mass_frac_NaCl * 1e6))

    prod_flow_mass = sum(
        m.fs.product.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    prod_mass_frac_NaCl = (
        m.fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"].value / prod_flow_mass
    )
    print("Product: %.3f kg/s, %.0f ppm" % (prod_flow_mass, prod_mass_frac_NaCl * 1e6))

    print(
        "Volumetric recovery: %.1f%%"
        % (value(m.fs.RO.recovery_vol_phase[0, "Liq"]) * 100)
    )
    print(
        "Water recovery: %.1f%%"
        % (value(m.fs.RO.recovery_mass_phase_comp[0, "Liq", "H2O"]) * 100)
    )
    print(
        "Energy Consumption: %.1f kWh/m3"
        % value(m.fs.costing.specific_energy_consumption)
    )
    print("Levelized cost of water: %.2f $/m3" % value(m.fs.costing.LCOW))


def display_design(m):
    print("---decision variables---")
    print("Operating pressure %.1f bar" % (m.fs.RO.inlet.pressure[0].value / 1e5))
    print("Membrane area %.1f m2" % (m.fs.RO.area.value))

    print("---design variables---")
    print(
        "Pump 1\noutlet pressure: %.1f bar\npower %.2f kW"
        % (
            m.fs.P1.outlet.pressure[0].value / 1e5,
            m.fs.P1.work_mechanical[0].value / 1e3,
        )
    )
    if m.fs.erd_type == ERDtype.pressure_exchanger:
        print("Separator")
        print("Split fraction %.2f" % (m.fs.S1.split_fraction[0, "PXR"].value * 100))
        print(
            "Pump 2\noutlet pressure: %.1f bar\npower %.2f kW"
            % (
                m.fs.P2.outlet.pressure[0].value / 1e5,
                m.fs.P2.work_mechanical[0].value / 1e3,
            )
        )
    elif m.fs.erd_type == ERDtype.pump_as_turbine:
        print(
            "ERD\ninlet pressure: %.1f bar\npower recovered %.2f kW"
            % (
                m.fs.ERD.inlet.pressure[0].value / 1e5,
                -1 * m.fs.ERD.work_mechanical[0].value / 1e3,
            )
        )
    else:
        erd_type_not_found(m.fs.erd_type)


def display_state(m):
    print("---state---")

    def print_state(s, b):
        flow_mass = sum(
            b.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
        )
        mass_frac_ppm = b.flow_mass_phase_comp[0, "Liq", "NaCl"].value / flow_mass * 1e6
        pressure_bar = b.pressure[0].value / 1e5
        print(
            s
            + ": %.3f kg/s, %.0f ppm, %.1f bar"
            % (flow_mass, mass_frac_ppm, pressure_bar)
        )

    print_state("Feed      ", m.fs.feed.outlet)
    print_state("Split 1   ", m.fs.S1.P1)
    print_state("P1 out    ", m.fs.P1.outlet)
    print_state("Split 2   ", m.fs.S1.PXR)
    print_state("PXR LP out", m.fs.PXR.low_pressure_outlet)
    print_state("P2 out    ", m.fs.P2.outlet)
    print_state("Mix out   ", m.fs.M1.outlet)
    print_state("RO perm   ", m.fs.RO.permeate)
    print_state("RO reten  ", m.fs.RO.retentate)
    print_state("PXR HP out", m.fs.PXR.high_pressure_outlet)


if __name__ == "__main__":
    # m = main(erd_type=ERDtype.pressure_exchanger)
    m = main(erd_type=ERDtype.pump_as_turbine)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Block,
    Var,
    Suffix,
    units as pyunits,
    ExternalFunction,
    check_optimal_termination,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
)
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.util.functions import functions_lib
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import InitializationMixin
from watertap.costing.unit_models.evaporator import cost_evaporator


_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Evaporator")
class EvaporatorData(InitializationMixin, UnitModelBlockData):
    """
    Evaporator model for MVC
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package_feed",
        ConfigValue(
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args_feed",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "property_package_vapor",
        ConfigValue(
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args_vapor",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
        and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
        **default** - MaterialBalanceType.useDefault.
        **Valid values:** {
        **MaterialBalanceType.useDefault - refer to property package for default
        balance type
        **MaterialBalanceType.none** - exclude material balances,
        **MaterialBalanceType.componentPhase** - use phase component balances,
        **MaterialBalanceType.componentTotal** - use total component balances,
        **MaterialBalanceType.elementTotal** - use total element balances,
        **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
        **default** - EnergyBalanceType.useDefault.
        **Valid values:** {
        **EnergyBalanceType.useDefault - refer to property package for default
        balance type
        **EnergyBalanceType.none** - exclude energy balances,
        **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
        **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
        **EnergyBalanceType.energyTotal** - single energy balance for material,
        **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    def build(self):
        super().build()

        if self.config.property_package_feed is None:
            raise ConfigurationError(
                "Users must provide a feed property package to the evaporator unit model"
            )
        if self.config.property_package_vapor is None:
            raise ConfigurationError(
                "Users must provide a vapor property package to the evaporator unit model"
            )

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta_feed = (
            self.config.property_package_feed.get_metadata().get_derived_units
        )

        # Add shared unit model variables
        self.U = Var(
            initialize=1e3,
            bounds=(10, 1e4),
            units=pyunits.J * pyunits.s**-1 * pyunits.m**-2 * pyunits.K**-1,
        )

        self.area = Var(initialize=1e2, bounds=(1e-1, 1e4), units=pyunits.m**2)

        self.delta_temperature_in = Var(
            initialize=1e1, bounds=(0.0, 1e3), units=pyunits.K
        )

        self.delta_temperature_out = Var(
            initialize=1e1, bounds=(0.0, 1e3), units=pyunits.K
        )

        self.lmtd = Var(initialize=1e1, bounds=(0.0, 1e3), units=pyunits.K)

        self.heat_transfer = Var(
            initialize=1e4, bounds=(1, 1e10), units=pyunits.J * pyunits.s**-1
        )

        # Add state blocks
        # Feed state block
        tmp_dict = dict(**self.config.property_package_args_feed)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package_feed
        tmp_dict["defined_state"] = True  # feed inlet defined
        self.properties_feed = self.config.property_package_feed.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of feed inlet",
            **tmp_dict,
        )

        # Brine state block
        tmp_dict["defined_state"] = False  # brine outlet not yet defined
        self.properties_brine = self.config.property_package_feed.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of brine outlet",
            **tmp_dict,
        )

        # Vapor state block
        tmp_dict = dict(**self.config.property_package_args_vapor)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package_vapor
        tmp_dict["defined_state"] = False  # vapor outlet not yet defined
        self.properties_vapor = self.config.property_package_vapor.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of vapor outlet",
            **tmp_dict,
        )

        # Add block for condenser constraints
        self.connection_to_condenser = Block()

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name="inlet_feed", block=self.properties_feed)
        self.add_port(name="outlet_brine", block=self.properties_brine)
        self.add_port(name="outlet_vapor", block=self.properties_vapor)

        # Mass balance
        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package_feed.component_list,
            doc="Mass balance",
        )
        def eq_mass_balance(b, t, j):
            lb = b.properties_vapor[t].flow_mass_phase_comp["Liq", "H2O"].lb
            b.properties_vapor[t].flow_mass_phase_comp["Liq", "H2O"].fix(lb)
            if j == "H2O":
                return (
                    b.properties_feed[t].flow_mass_phase_comp["Liq", "H2O"]
                    == b.properties_brine[t].flow_mass_phase_comp["Liq", "H2O"]
                    + b.properties_vapor[t].flow_mass_phase_comp["Vap", "H2O"]
                )
            else:
                return (
                    b.properties_feed[t].flow_mass_phase_comp["Liq", j]
                    == b.properties_brine[t].flow_mass_phase_comp["Liq", j]
                )

        # Energy balance
        @self.Constraint(self.flowsheet().time, doc="Energy balance")
        def eq_energy_balance(b, t):
            return (
                b.heat_transfer + b.properties_feed[t].enth_flow
                == b.properties_brine[t].enth_flow
                + b.properties_vapor[t].enth_flow_phase["Vap"]
            )

        # Brine pressure
        @self.Constraint(self.flowsheet().time, doc="Brine pressure")
        def eq_brine_pressure(b, t):
            return b.properties_brine[t].pressure == b.properties_brine[t].pressure_sat

        # Vapor pressure
        @self.Constraint(self.flowsheet().time, doc="Vapor pressure")
        def eq_vapor_pressure(b, t):
            return b.properties_vapor[t].pressure == b.properties_brine[t].pressure

        # Vapor temperature - assumed to be equal to brine temperature
        @self.Constraint(self.flowsheet().time, doc="Vapor temperature")
        def eq_vapor_temperature(b, t):
            return (
                b.properties_vapor[t].temperature == b.properties_brine[t].temperature
            )

        ### EVAPORATOR CONSTRAINTS ###
        # log mean temperature
        @self.Constraint(self.flowsheet().time, doc="Log mean temperature difference")
        def eq_lmtd(b, t):
            dT_in = b.delta_temperature_in
            dT_out = b.delta_temperature_out
            temp_units = pyunits.get_units(dT_in)
            dT_avg = (dT_in + dT_out) / 2
            # external function that ruturns the real root, for the cuberoot of negitive
            # numbers, so it will return without error for positive and negitive dT.
            b.cbrt = ExternalFunction(
                library=functions_lib(), function="cbrt", arg_units=[temp_units**3]
            )
            return b.lmtd == b.cbrt((dT_in * dT_out * dT_avg)) * temp_units

        # Evaporator heat transfer
        @self.Constraint(self.flowsheet().time, doc="Evaporator heat transfer")
        def eq_evaporator_heat(b, t):
            return b.heat_transfer == b.U * b.area * b.lmtd

    def connect_to_condenser(self, condenser_blk):
        # Temperature difference in
        @self.connection_to_condenser.Constraint(
            self.flowsheet().time, doc="Temperature difference in"
        )
        def eq_delta_temperature_in(b, t):
            return (
                self.delta_temperature_in
                == condenser_blk.control_volume.properties_in[t].temperature
                - self.properties_brine[t].temperature
            )

        # Temperature difference out
        @self.connection_to_condenser.Constraint(
            self.flowsheet().time, doc="Temperature difference out"
        )
        def eq_delta_temperature_out(b, t):
            return (
                self.delta_temperature_out
                == condenser_blk.control_volume.properties_out[t].temperature
                - self.properties_brine[t].temperature
            )

        # Heat transfer between feed side and condenser
        @self.connection_to_condenser.Constraint(
            self.flowsheet().time, doc="Heat transfer balance"
        )
        def eq_heat_balance(b, t):
            return self.heat_transfer == -condenser_blk.control_volume.heat[t]

    def initialize_build(
        self,
        delta_temperature_in=None,
        delta_temperature_out=None,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            delta_temperature_in : value to fix delta_temperature_in
            delta_temperature_out : value to fix delta_temperature_out
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        if hasattr(self, "connection_to_condenser"):
            self.connection_to_condenser.deactivate()

        # ---------------------------------------------------------------------
        # Initialize feed side
        flags_feed = self.properties_feed.initialize(
            solver=solver, optarg=optarg, hold_state=True
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # # ---------------------------------------------------------------------
        # # Initialize brine
        # Set state_args from inlet state
        if state_args is None:
            state_args = {}
            state_dict = self.properties_feed[
                self.flowsheet().config.time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        self.properties_brine.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args
        )

        state_args_vapor = {}
        state_args_vapor["pressure"] = 0.5 * state_args["pressure"]
        state_args_vapor["temperature"] = state_args["temperature"]
        state_args_vapor["flow_mass_phase_comp"] = {
            ("Liq", "H2O"): self.properties_vapor[0]
            .flow_mass_phase_comp["Liq", "H2O"]
            .lb,
            ("Vap", "H2O"): state_args["flow_mass_phase_comp"][("Liq", "H2O")],
        }

        self.properties_vapor.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_vapor,
        )

        init_log.info_high("Initialization Step 2 Complete.")

        # incorporate guessed temperature differences
        has_guessed_delta_temperature_in = False
        if delta_temperature_in is not None:
            if self.delta_temperature_in.is_fixed():
                raise RuntimeError(
                    "A guess was provided for the delta_temperature_in variable in the "
                    "initialization, but it is already fixed. Either do not "
                    "provide a guess for or unfix delta_temperature_in"
                )
            self.delta_temperature_in.fix(delta_temperature_in)
            has_guessed_delta_temperature_in = True

        has_guessed_delta_temperature_out = False
        if delta_temperature_out is not None:
            if self.delta_temperature_out.is_fixed():
                raise RuntimeError(
                    "A guess was provided for the delta_temperature_out variable in the "
                    "initialization, but it is already fixed. Either do not "
                    "provide a guess for or unfix delta_temperature_out"
                )
            self.delta_temperature_out.fix(delta_temperature_out)
            has_guessed_delta_temperature_out = True

        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release feed and condenser inlet states and release delta_temperature
        self.properties_feed.release_state(flags_feed, outlvl=outlvl)
        if has_guessed_delta_temperature_in:
            self.delta_temperature_in.unfix()
        if has_guessed_delta_temperature_out:
            self.delta_temperature_out.unfix()
        if hasattr(self, "connection_to_condenser"):
            self.connection_to_condenser.activate()

        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    def _get_performance_contents(self, time_point=0):
        var_dict = {
            "Heat transfer": self.heat_transfer,
            "Evaporator temperature": self.properties_brine[0].temperature,
            "Evaporator pressure": self.properties_brine[0].pressure,
        }

        return {"vars": var_dict}

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        if iscale.get_scaling_factor(self.heat_transfer) is None:
            sf = iscale.get_scaling_factor(
                self.properties_vapor[0].enth_flow_phase["Vap"]
            )
            iscale.set_scaling_factor(self.heat_transfer, sf)

    @property
    def default_costing_method(self):
        return cost_evaporator



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order EC model
"""
import pytest


from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Param,
    value,
    Var,
    Block,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock
import idaes.core.util.scaling as iscale

from watertap.unit_models.zero_order import ElectrocoagulationZO
from watertap.unit_models.zero_order.electrocoagulation_zo import (
    ElectrodeMaterial,
    ReactorMaterial,
    OverpotentialCalculation,
)
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestECZO_AL:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "tds",
                "tss",
                "toc",
            ]
        )

        m.fs.unit = ElectrocoagulationZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(43.8)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(0.004599)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(0.5527998)
        m.fs.unit.inlet.flow_mass_comp[0, "tds"].fix(5.256)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.config.electrode_material, ElectrodeMaterial)
        assert isinstance(model.fs.unit.config.reactor_material, ReactorMaterial)
        assert isinstance(
            model.fs.unit.config.overpotential_calculation, OverpotentialCalculation
        )
        assert model.fs.unit.config.electrode_material == ElectrodeMaterial.aluminum
        assert model.fs.unit.config.reactor_material == ReactorMaterial.pvc
        assert (
            model.fs.unit.config.overpotential_calculation
            == OverpotentialCalculation.fixed
        )
        assert value(model.fs.unit.mw_electrode_material) == 0.027
        assert value(model.fs.unit.valence_electrode_material) == 3
        assert value(model.fs.unit.density_electrode_material) == 2710
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == "electrocoagulation"
        assert isinstance(model.fs.unit.mw_electrode_material, Param)
        assert isinstance(model.fs.unit.valence_electrode_material, Param)
        assert isinstance(model.fs.unit.density_electrode_material, Param)
        assert isinstance(model.fs.unit.recovery_frac_mass_H2O, Var)
        assert isinstance(model.fs.unit.removal_frac_mass_comp, Var)
        assert isinstance(model.fs.unit.power_required, Var)
        assert isinstance(model.fs.unit.overpotential, Var)
        assert isinstance(model.fs.unit.ohmic_resistance, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("electrocoagulation")
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.8
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.99

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"]:
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_scaling(self, model):
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("H2O"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("tds"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("tss"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e4, index=("toc"))
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["tds"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["tss"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["toc"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["H2O"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["tss"], 1e3
        )
        iscale.calculate_scaling_factors(model)
        badly_scaled_var_list = list(iscale.badly_scaled_var_generator(model))
        assert len(badly_scaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(43.3619999, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["H2O"]
        )
        assert pytest.approx(1.5768, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["tds"]
        )
        assert pytest.approx(0.165839, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["tss"]
        )
        assert pytest.approx(0.00137970000, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["toc"]
        )
        assert pytest.approx(53188.5028, rel=1e-2) == value(
            model.fs.unit.applied_current
        )
        assert pytest.approx(2, rel=1e-2) == value(model.fs.unit.cell_voltage)
        assert pytest.approx(9.4e-6, rel=1e-2) == value(model.fs.unit.ohmic_resistance)
        assert pytest.approx(106377, rel=1e-2) == value(model.fs.unit.power_required)

    @pytest.mark.component
    def test_costing(self, model):
        m = model
        ec = m.fs.unit
        m.fs.costing = ZeroOrderCosting()
        m.fs.costing.base_currency = pyunits.USD_2018
        ec.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(ec.properties_treated[0].flow_vol)
        m.fs.costing.add_electricity_intensity(ec.properties_treated[0].flow_vol)
        assert "aluminum" in m.fs.costing._registered_flows
        assert (
            value(m.fs.costing.electrocoagulation.ec_reactor_cap_material_coeff[None])
            == 0.062
        )
        assert (
            value(m.fs.costing.electrocoagulation.electrode_material_cost[None]) == 2.23
        )
        assert isinstance(m.fs.costing.electrocoagulation, Block)
        assert isinstance(
            m.fs.costing.electrocoagulation.ec_power_supply_base_slope, Var
        )
        assert isinstance(m.fs.costing.electrocoagulation.ec_reactor_cap_base, Var)
        assert isinstance(m.fs.unit.costing.capital_cost, Var)
        assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

        assert degrees_of_freedom(m.fs.unit) == 0

        results = solver.solve(m)
        check_optimal_termination(results)
        assert pytest.approx(0.34090, rel=1e-3) == value(m.fs.costing.LCOW)
        assert pytest.approx(0.65510, rel=1e-3) == value(
            m.fs.costing.electricity_intensity
        )
        assert pytest.approx(4928.611, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_reactor
        )
        assert pytest.approx(55926.1017, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_power_supply
        )

        assert pytest.approx(13006.1652, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_electrodes
        )

        assert (
            ec.costing.electricity_flow in m.fs.costing._registered_flows["electricity"]
        )


class TestECZO_FE:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "tds",
                "tss",
                "toc",
            ]
        )

        m.fs.unit = ElectrocoagulationZO(
            property_package=m.fs.params,
            database=m.db,
            electrode_material="iron",
            reactor_material="stainless_steel",
        )

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(43.8)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(0.004599)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(0.5527998)
        m.fs.unit.inlet.flow_mass_comp[0, "tds"].fix(5.256)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.config.electrode_material, ElectrodeMaterial)
        assert isinstance(model.fs.unit.config.reactor_material, ReactorMaterial)
        assert isinstance(
            model.fs.unit.config.overpotential_calculation, OverpotentialCalculation
        )
        assert model.fs.unit.config.electrode_material == ElectrodeMaterial.iron
        assert model.fs.unit.config.reactor_material == ReactorMaterial.stainless_steel
        assert (
            model.fs.unit.config.overpotential_calculation
            == OverpotentialCalculation.fixed
        )
        assert value(model.fs.unit.mw_electrode_material) == 0.056
        assert value(model.fs.unit.valence_electrode_material) == 2
        assert value(model.fs.unit.density_electrode_material) == 7860
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == "electrocoagulation"
        assert isinstance(model.fs.unit.mw_electrode_material, Param)
        assert isinstance(model.fs.unit.valence_electrode_material, Param)
        assert isinstance(model.fs.unit.density_electrode_material, Param)
        assert isinstance(model.fs.unit.recovery_frac_mass_H2O, Var)
        assert isinstance(model.fs.unit.removal_frac_mass_comp, Var)
        assert isinstance(model.fs.unit.power_required, Var)
        assert isinstance(model.fs.unit.overpotential, Var)
        assert isinstance(model.fs.unit.ohmic_resistance, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("electrocoagulation")
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.8
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.99

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"]:
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_scaling(self, model):
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("H2O"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("tds"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("tss"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e4, index=("toc"))
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["tds"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["tss"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["toc"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["H2O"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["tss"], 1e3
        )
        iscale.calculate_scaling_factors(model)
        badly_scaled_var_list = list(iscale.badly_scaled_var_generator(model))
        assert len(badly_scaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(43.3619999, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["H2O"]
        )
        assert pytest.approx(1.5768, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["tds"]
        )
        assert pytest.approx(0.165839, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["tss"]
        )
        assert pytest.approx(0.00137970000, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["toc"]
        )
        assert pytest.approx(17096.3045, rel=1e-2) == value(
            model.fs.unit.applied_current
        )
        assert pytest.approx(2, rel=1e-2) == value(model.fs.unit.cell_voltage)
        assert pytest.approx(2.925e-5, rel=1e-2) == value(
            model.fs.unit.ohmic_resistance
        )
        assert pytest.approx(34192.609, rel=1e-2) == value(model.fs.unit.power_required)

    @pytest.mark.component
    def test_costing(self, model):

        m = model
        ec = m.fs.unit
        m.fs.costing = ZeroOrderCosting()
        m.fs.costing.base_currency = pyunits.USD_2018
        ec.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(ec.properties_treated[0].flow_vol)
        m.fs.costing.add_electricity_intensity(ec.properties_treated[0].flow_vol)
        assert "iron" in m.fs.costing._registered_flows
        assert (
            value(m.fs.costing.electrocoagulation.ec_reactor_cap_material_coeff[None])
            == 3.4
        )
        assert (
            value(m.fs.costing.electrocoagulation.electrode_material_cost[None]) == 3.41
        )
        assert isinstance(m.fs.costing.electrocoagulation, Block)
        assert isinstance(
            m.fs.costing.electrocoagulation.ec_power_supply_base_slope, Var
        )
        assert isinstance(m.fs.costing.electrocoagulation.ec_reactor_cap_base, Var)
        assert isinstance(m.fs.unit.costing.capital_cost, Var)
        assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

        assert degrees_of_freedom(m.fs.unit) == 0

        results = solver.solve(m)
        check_optimal_termination(results)
        assert pytest.approx(0.4696, rel=1e-3) == value(m.fs.costing.LCOW)
        assert pytest.approx(0.21057, rel=1e-3) == value(
            m.fs.costing.electricity_intensity
        )
        assert pytest.approx(162180.930, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_reactor
        )
        assert pytest.approx(17976.2465, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_power_supply
        )

        assert pytest.approx(18541.1436, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_electrodes
        )

        assert (
            ec.costing.electricity_flow in m.fs.costing._registered_flows["electricity"]
        )


class TestECZO_OverpotentialCalculation:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "tds",
                "tss",
                "toc",
            ]
        )

        m.fs.unit = ElectrocoagulationZO(
            property_package=m.fs.params,
            database=m.db,
            reactor_material="stainless_steel",
            overpotential_calculation="calculated",
        )

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(43.8)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(0.004599)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(0.5527998)
        m.fs.unit.inlet.flow_mass_comp[0, "tds"].fix(5.256)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.config.electrode_material, ElectrodeMaterial)
        assert isinstance(model.fs.unit.config.reactor_material, ReactorMaterial)
        assert isinstance(
            model.fs.unit.config.overpotential_calculation, OverpotentialCalculation
        )
        assert model.fs.unit.config.electrode_material == ElectrodeMaterial.aluminum
        assert model.fs.unit.config.reactor_material == ReactorMaterial.stainless_steel
        assert (
            model.fs.unit.config.overpotential_calculation
            == OverpotentialCalculation.calculated
        )
        assert value(model.fs.unit.mw_electrode_material) == 0.027
        assert value(model.fs.unit.valence_electrode_material) == 3
        assert value(model.fs.unit.density_electrode_material) == 2710
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == "electrocoagulation"
        assert isinstance(model.fs.unit.mw_electrode_material, Param)
        assert isinstance(model.fs.unit.valence_electrode_material, Param)
        assert isinstance(model.fs.unit.density_electrode_material, Param)
        assert isinstance(model.fs.unit.recovery_frac_mass_H2O, Var)
        assert isinstance(model.fs.unit.removal_frac_mass_comp, Var)
        assert isinstance(model.fs.unit.power_required, Var)
        assert isinstance(model.fs.unit.overpotential, Var)
        assert isinstance(model.fs.unit.ohmic_resistance, Var)
        assert isinstance(model.fs.unit.overpotential_k1, Var)
        assert isinstance(model.fs.unit.overpotential_k2, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("electrocoagulation")
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.8
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.99

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"]:
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_scaling(self, model):
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("H2O"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("tds"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e-3, index=("tss"))
        model.fs.params.set_default_scaling("flow_mass_comp", 1e4, index=("toc"))
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["tds"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["tss"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_treated[0].flow_mass_comp["toc"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["H2O"], 1e3
        )
        iscale.set_scaling_factor(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["tss"], 1e3
        )
        iscale.calculate_scaling_factors(model)
        badly_scaled_var_list = list(iscale.badly_scaled_var_generator(model))
        assert len(badly_scaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(43.3619999, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["H2O"]
        )
        assert pytest.approx(1.5768, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["tds"]
        )
        assert pytest.approx(0.165839, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["tss"]
        )
        assert pytest.approx(0.00137970000, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_mass_comp["toc"]
        )
        assert pytest.approx(53188.5028, rel=1e-2) == value(
            model.fs.unit.applied_current
        )
        assert pytest.approx(2.49011, rel=1e-2) == value(model.fs.unit.cell_voltage)
        assert pytest.approx(9.4e-6, rel=1e-2) == value(model.fs.unit.ohmic_resistance)
        assert pytest.approx(132445, rel=1e-2) == value(model.fs.unit.power_required)

    @pytest.mark.component
    def test_costing(self, model):

        m = model
        ec = m.fs.unit
        m.fs.costing = ZeroOrderCosting()
        m.fs.costing.base_currency = pyunits.USD_2018
        ec.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(ec.properties_treated[0].flow_vol)
        m.fs.costing.add_electricity_intensity(ec.properties_treated[0].flow_vol)
        assert "aluminum" in m.fs.costing._registered_flows
        assert (
            value(m.fs.costing.electrocoagulation.ec_reactor_cap_material_coeff[None])
            == 3.4
        )
        assert (
            value(m.fs.costing.electrocoagulation.electrode_material_cost[None]) == 2.23
        )
        assert isinstance(m.fs.costing.electrocoagulation, Block)
        assert isinstance(
            m.fs.costing.electrocoagulation.ec_power_supply_base_slope, Var
        )
        assert isinstance(m.fs.costing.electrocoagulation.ec_reactor_cap_base, Var)
        assert isinstance(m.fs.unit.costing.capital_cost, Var)
        assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

        assert degrees_of_freedom(m.fs.unit) == 0

        results = solver.solve(m)
        check_optimal_termination(results)
        assert pytest.approx(0.406240, rel=1e-3) == value(m.fs.costing.LCOW)
        assert pytest.approx(0.81564, rel=1e-3) == value(
            m.fs.costing.electricity_intensity
        )
        assert pytest.approx(270278.669, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_reactor
        )
        assert pytest.approx(69631.117, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_power_supply
        )

        assert pytest.approx(13006.16527, rel=1e-3) == value(
            m.fs.unit.costing.capital_cost_electrodes
        )

        assert (
            ec.costing.electricity_flow in m.fs.costing._registered_flows["electricity"]
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from watertap.tools.parameter_sweep.sampling_types import (
    LinearSample,
    GeomSample,
    ReverseGeomSample,
    UniformSample,
    NormalSample,
    LatinHypercubeSample,
    PredeterminedFixedSample,
    PredeterminedRandomSample,
)
import yaml
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


class ParameterSweepReader:
    @staticmethod
    def _yaml_to_dict(yaml_filename):
        """Reads and stores a yaml file as a dictionary

        Args:
            yaml_filename (str):
                The filename of the yaml file to read.

        Returns:
            input_dict (dict):
                The result of reading the yaml file and translating
                its structure into a dictionary.

        """

        try:
            # Open the yaml file and import the contents into a
            # dictionary with the same structure
            with open(yaml_filename) as fp:
                input_dict = yaml.load(fp, Loader=yaml.FullLoader)

        except:
            raise ValueError("Could not open file %s" % (yaml_filename))

        return input_dict

    def get_sweep_params_from_yaml(self, m, yaml_filename):
        """Creates a dictionary of swept model parameters specified via yaml file

        This function creates a dictionary of the items to vary during a parameter
        sweep where the variable name, model attribute, and sweeping domain are
        specified in a YAML file.  The YAML file should have the following format::

            A_comp:
                type: NormalSample
                param: fs.RO.A_comp
                mean: 4.0e-12
                std: 0.5e-12

        where the top-level keyword can be any short, easily understood identifier
        for the parameter.  ``type`` must be one of ``LinearSample``, ``UniformSample``,
        ``NormalSample``, or ``LatinHypercubeSample``.  ``param`` must be a valid
        dot-sperated string path to the object attribute (in this case, an RO attribute
        on the flowsheet ``m``) that you wish to vary.  The remaining arguments are
        dependent on the sample type selected.  For ``NormalSample`` information about
        the mean and standard deviation is required.  Consult the ``parameter_sweep``
        help for more information on the different sample classes.

        Args:
            m (pyomo model):
                The flowsheet containing the model to deploy with the parameter sweep
                tool.
            yaml_filename (str):
                The path to the yaml file.

        Returns:
            sweep_params (dict):
                A dictionary containing different instances of parameter sweep samples

        """

        input_dict = self._yaml_to_dict(yaml_filename)
        return self._dict_to_params(m, input_dict)

    @staticmethod
    def _dict_to_diff_spec(m, input_dict):
        """Reads and stores a yaml file as a dictionary
        Args:
            dict (str):
                The dictionary of paramters that are turned into paramter sweep samples
        Returns:
            input_dict (dict):
                The result of reading the yaml file and translating
                its structure into a dictionary.
        """
        diff_spec = {}

        for param, values in input_dict.items():
            # Find the specified component on the model
            component = m.find_component(values["param"])

            if component is None:
                raise ValueError(f'Could not acccess attribute {values["param"]}')

            if values["diff_sample_type"] == "NormalSample":
                diff_spec[param] = {
                    "diff_mode": values["diff_mode"],
                    "diff_sample_type": NormalSample,
                    "std_dev": values["std_dev"],
                    "pyomo_object": component,
                }

            elif values["diff_sample_type"] == "UniformSample":
                if values["diff_mode"] == "percentile":
                    nominal_lb = values["nominal_lb"]
                    nominal_ub = values["nominal_ub"]
                elif values["diff_mode"] == "sum" or values["diff_mode"] == "product":
                    nominal_lb = None
                    nominal_ub = None
                diff_spec[param] = {
                    "diff_mode": values["diff_mode"],
                    "diff_sample_type": UniformSample,
                    "relative_lb": values["relative_lb"],
                    "relative_ub": values["relative_ub"],
                    "nominal_lb": nominal_lb,
                    "nominal_ub": nominal_ub,
                    "pyomo_object": component,
                }
            elif values["diff_sample_type"] == "LinearSample":
                if values["diff_mode"] == "percentile":
                    nominal_lb = values["nominal_lb"]
                    nominal_ub = values["nominal_ub"]
                elif values["diff_mode"] == "sum" or values["diff_mode"] == "product":
                    nominal_lb = None
                    nominal_ub = None
                diff_spec[param] = {
                    "diff_mode": values["diff_mode"],
                    "diff_sample_type": LinearSample,
                    "relative_lb": values["relative_lb"],
                    "relative_ub": values["relative_ub"],
                    "nominal_lb": nominal_lb,
                    "nominal_ub": nominal_ub,
                    "pyomo_object": component,
                }

        return diff_spec

    @staticmethod
    def _dict_to_params(m, input_dict):
        """Reads and stores a yaml file as a dictionary

        Args:
            dict (str):
                The dictionary of paramters that are turned into paramter sweep samples

        Returns:
            input_dict (dict):
                The result of reading the yaml file and translating
                its structure into a dictionary.

        """
        sweep_params = {}

        for param, values in input_dict.items():
            # Find the specified component on the model
            component = m.find_component(values["param"])

            if component is None:
                raise ValueError(f'Could not acccess attribute {values["param"]}')

            if values["type"] == "LinearSample":
                sweep_params[param] = LinearSample(
                    component,
                    values["lower_limit"],
                    values["upper_limit"],
                    values["num_samples"],
                )
            elif values["type"] == "GeomSample":
                sweep_params[param] = GeomSample(
                    component,
                    values["lower_limit"],
                    values["upper_limit"],
                    values["num_samples"],
                )
            elif values["type"] == "ReverseGeomSample":
                sweep_params[param] = ReverseGeomSample(
                    component,
                    values["lower_limit"],
                    values["upper_limit"],
                    values["num_samples"],
                )
            elif values["type"] == "UniformSample":
                sweep_params[param] = UniformSample(
                    component,
                    values["lower_limit"],
                    values["upper_limit"],
                    values["num_samples"],
                )

            elif values["type"] == "NormalSample":
                sweep_params[param] = NormalSample(
                    component,
                    values["mean"],
                    values["std"],
                    values["num_samples"],
                )

            elif values["type"] == "LatinHypercubeSample":
                sweep_params[param] = LatinHypercubeSample(
                    component,
                    values["lower_limit"],
                    values["upper_limit"],
                    values["num_samples"],
                )

            elif values["type"] == "PredeterminedFixedSample":
                sweep_params[param] = PredeterminedFixedSample(
                    component, values["array"]
                )
            elif values["type"] == "PredeterminedRandomSample":
                sweep_params[param] = PredeterminedRandomSample(
                    component, values["array"]
                )
        return sweep_params

    @staticmethod
    def _set_value(component, key, default_value):
        _log.debug(f"Property: {key}")
        _log.debug(f"New Value: {default_value:.6e}, Old Value: {component.value:.6e}")
        component.value = default_value

    def set_defaults_from_yaml(self, m, yaml_filename, verbose=False):
        """Sets default model values using values stored in a yaml file

        This function reads a yaml file with the structure::

            fs.path.to.attribute_1: 0.123
            fs.path.to.attribute_2: 1.234
            ...

        and uses the (key, default_value) pairs to set default values
        for the attributes in model ``m``.

        Args:
            m (pyomo model):
                The flowsheet containing the model to set default values for
            yaml_filename (str):
                The path to the yaml file.

        Returns:
            N/A

        """

        input_dict = self._yaml_to_dict(yaml_filename)
        self._set_values_from_dict(m, input_dict, verbose)

    def _set_values_from_dict(self, m, input_dict, verbose=False):
        fail_count = 0

        for key, default_value in input_dict.items():
            # Find the specified component on the model
            component = m.find_component(key)

            if component is None:
                raise ValueError(f"Could not acccess component {key} on model {m}")

            if component.is_variable_type():
                self._set_value(component, key, default_value)
            elif component.is_parameter_type():
                if component.mutable:
                    self._set_value(component, key, default_value)
                else:
                    _log.warning(f"Cannot set value of non-mutable Param {component}")
                    fail_count += 1
            else:
                _log.warning(f"Cannot set value of component {component}")
                fail_count += 1

        number_defaults = len(input_dict)
        print(
            f"Set {number_defaults-fail_count} of {number_defaults} options to default values ({fail_count} failures, see warnings)"
        )


def get_sweep_params_from_yaml(m, yaml_filename):
    """Creates a dictionary of swept model parameters specified via yaml file

    This function creates a dictionary of the items to vary during a parameter
    sweep where the variable name, model attribute, and sweeping domain are
    specified in a YAML file.  The YAML file should have the following format::

        A_comp:
            type: NormalSample
            param: fs.RO.A_comp
            mean: 4.0e-12
            std: 0.5e-12

    where the top-level keyword can be any short, easily understood identifier
    for the parameter.  ``type`` must be one of ``LinearSample``, ``UniformSample``,
    ``NormalSample``, or ``LatinHypercubeSample``.  ``param`` must be a valid
    dot-sperated string path to the object attribute (in this case, an RO attribute
    on the flowsheet ``m``) that you wish to vary.  The remaining arguments are
    dependent on the sample type selected.  For ``NormalSample`` information about
    the mean and standard deviation is required.  Consult the ``parameter_sweep``
    help for more information on the different sample classes.

    Args:
        m (pyomo model):
            The flowsheet containing the model to deploy with the parameter sweep
            tool.
        yaml_filename (str):
            The path to the yaml file.

    Returns:
        sweep_params (dict):
            A dictionary containing different instances of parameter sweep samples

    """
    return ParameterSweepReader().get_sweep_params_from_yaml(m, yaml_filename)


def set_defaults_from_yaml(m, yaml_filename, verbose=False):
    """Sets default model values using values stored in a yaml file

    This function reads a yaml file with the structure::

        fs.path.to.attribute_1: 0.123
        fs.path.to.attribute_2: 1.234
        ...

    and uses the (key, default_value) pairs to set default values
    for the attributes in model ``m``.

    Args:
        m (pyomo model):
            The flowsheet containing the model to set default values for
        yaml_filename (str):
            The path to the yaml file.

    Returns:
        N/A

    """
    return ParameterSweepReader().set_defaults_from_yaml(m, yaml_filename, verbose)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a deep well injection unit.
"""

import pyomo.environ as pyo
from pyomo.environ import Reference, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_pt, pump_electricity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("DeepWellInjectionZO")
class DeepWellInjectionZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a deep well injection unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "deep_well_injection"
        build_pt(self)
        self._Q = Reference(self.properties[:].flow_vol)

        pump_electricity(self, self._Q)

        self.pipe_distance = Var(
            self.flowsheet().config.time, units=pyunits.miles, doc="Piping distance"
        )

        self.pipe_diameter = Var(
            self.flowsheet().config.time, units=pyunits.inches, doc="Pipe diameter"
        )

        self.flow_basis = Var(
            self.flowsheet().time, units=pyunits.m**3 / pyunits.hour, doc="flow basis"
        )

        self._fixed_perf_vars.append(self.pipe_distance)
        self._fixed_perf_vars.append(self.pipe_diameter)
        self._fixed_perf_vars.append(self.flow_basis)

        self._perf_var_dict["Pipe Distance (miles)"] = self.pipe_distance
        self._perf_var_dict["Pipe Diameter (inches)"] = self.pipe_diameter

    @property
    def default_costing_method(self):
        return self.cost_deep_well_injection

    @staticmethod
    def cost_deep_well_injection(blk, number_of_parallel_units=1):
        """
        General method for costing deep well injection processes. Capital cost
        is based on the cost of pump and pipe.
        This method also registers the electricity demand as a costed flow.
        Args:
            number_of_parallel_units (int, optional) - cost this unit as
                        number_of_parallel_units parallel units (default: 1)
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, C = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["well_pump_cost", "pipe_cost_basis", "flow_exponent"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        cost_well_pump = A

        cost_pipe = (
            B * blk.unit_model.pipe_distance[t0] * blk.unit_model.pipe_diameter[t0]
        )

        cost_total = pyo.units.convert(
            cost_well_pump + cost_pipe,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        Q = pyo.units.convert(
            blk.unit_model.properties[t0].flow_vol,
            to_units=pyo.units.m**3 / pyo.units.hour,
        )

        sizing_term = Q / blk.unit_model.flow_basis[t0]

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        blk.unit_model._general_power_law_form(
            blk,
            cost_total,
            C,
            sizing_term,
            factor,
            number_of_parallel_units,
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import value
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    property_models,
)


@pytest.mark.component
def test_feed_TDS():
    m = property_models.solve_specify_feed(base="TDS")
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "TDS"]) == pytest.approx(
        0.035, rel=1e-3
    )


@pytest.mark.component
def test_feed_ion():
    m = property_models.solve_specify_feed(base="ion")
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "Ca"]) == pytest.approx(
        3.82e-4, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "Cl"]) == pytest.approx(
        2.032e-2, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "H2O"]) == pytest.approx(
        0.9646, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "Mg"]) == pytest.approx(
        1.394e-3, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "Na"]) == pytest.approx(
        1.112e-2, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "SO4"]) == pytest.approx(
        2.136e-3, rel=1e-3
    )


@pytest.mark.component
def test_feed_salt():
    m = property_models.solve_specify_feed(base="salt")
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "CaSO4"]) == pytest.approx(
        1.298e-3, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "H2O"]) == pytest.approx(
        0.9646, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "MgCl2"]) == pytest.approx(
        4.251e-3, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "MgSO4"]) == pytest.approx(
        1.529e-3, rel=1e-3
    )
    assert value(m.fs.stream[0].mass_frac_phase_comp["Liq", "NaCl"]) == pytest.approx(
        2.827e-2, rel=1e-3
    )
